#include <linux/module.h>
#include<linux/init.h>
#include<linux/fs.h>
#include <linux/buffer_head.h>
#include "nngfs.h"

extern int nngfs_unlink(struct inode * dir, struct dentry *dentry);
extern int nngfs_create(struct inode *dir, struct dentry *dentry, int mode, struct nameidata *nd);
extern struct dentry *nngfs_lookup(struct inode *inode,struct dentry *de,struct nameidata *nd);
static inline __u32 ngfs_do_mod(void *a, __u32 b, int n)
{
 switch (n) {
  case 4:
   return *(__u32 *)a % b;
  case 8:
   {   
   unsigned long __upper, __low, __high, __mod;
   __u64 c = *(__u64 *)a;
   __upper = __high = c >> 32; 
   __low = c;
   if (__high) {
    __upper = __high % (b);
    __high = __high / (b);
   }   
   asm("divl %2":"=a" (__low), "=d" (__mod):"rm" (b), "0" (__low), "1" (__upper));
   asm("":"=A" (c):"a" (__low),"d" (__high));
   return __mod;
   }   
 }

 /* NOTREACHED */
 return 0;
}

#define do_mod(a, b) ngfs_do_mod(&(a), (b), sizeof(a))

static int nngfs_dir_read(struct file *filp,void *dirent,filldir_t filldir);

struct file_operations nngfs_dir_operations = {
        .llseek = generic_file_llseek,
        .read = generic_read_dir,
        .readdir = nngfs_dir_read,
};

int nngfs_dir_add(struct inode *dir,unsigned32 inum,const unsigned char *name){
  struct buffer_head *bh;
  struct nngfs_dir_entry de;
  struct nngfs_dir_entry *dip;
  unsigned32 blocknum = dir->i_ino+1;
  struct super_block  *sb = dir->i_sb;
  bh = sb_bread(sb,blocknum);
  if(!bh){
     printk(KERN_INFO "NGFS (%s) : Could not read block [%d] \n",__FUNCTION__,inum+1);
     return 1;
  }
  memset(&de,0,sizeof(de));
  de.inode = inum;
  de.name_len = strlen(name);
  strcpy(de.name,name);
  //de.rec_len = 3*sizeof(unsigned32) + de.name_len;
  de.rec_len = sizeof(struct nngfs_dir_entry);
  dip = (struct nngfs_dir_entry *)bh->b_data;
  /* Find a suitable place to insert the new data
   */
  while(dip->inode != NNGFS_INUM_INVALID && strcmp(dip->name,"") ){
      //   printk(KERN_INFO "name = [%s], inum [%d] \n",dip->name,dip->inode);
         dip = (struct nngfs_dir_entry *)((char *)dip + dip->rec_len);
  }
 // printk(KERN_INFO "name = [%s], inum [%d] \n",dip->name,dip->inode);
  memcpy(dip,&de,sizeof(de));
  mark_buffer_dirty(bh);
  brelse(bh);
  return 0;
}

static int nngfs_dir_read(struct file *file,void *dirent,filldir_t filldir){
      struct nngfs_dir_entry *nde = NULL;
      loff_t pos = file->f_pos;
      struct buffer_head *bh;
      struct inode *inode = file->f_path.dentry->d_inode;
      struct nngfs_inode *nino = (struct nngfs_inode *)(inode->i_private);
      struct super_block *sb = inode->i_sb;
      if(pos > inode->i_size)
	return 0; /* Got past the size mentioned in inode */
      bh = sb_bread(sb,nino->blockno);
      if(!bh){
#ifdef DEBUG 
      printk(KERN_INFO "NGFS (%s) : Could not read buffer\n",__FUNCTION__);
#endif
      return 0;
      }
      nde = (struct nngfs_dir_entry *)((char *)bh->b_data + do_mod(pos,sb->s_blocksize));
      if(!strcmp(nde->name,""))
         return 0;
      if(nde->inode != NNGFS_INUM_INVALID)
      {
          filldir(dirent,nde->name, nde->name_len, pos,nde->inode, DT_UNKNOWN);
          file->f_pos += (nde->rec_len); 
         // file->f_pos += sizeof(struct nngfs_dir_entry);
      }
      else{
          filldir(dirent,"", 0, pos,nde->inode, DT_UNKNOWN);
          file->f_pos += (nde->rec_len); 
         // file->f_pos += sizeof(struct nngfs_dir_entry);
      }
      brelse(bh);
      return 1;
}

unsigned32 clear_dentry(struct inode *nip ,int inum){
       struct super_block *sb = nip->i_sb;
       struct buffer_head *bh ;
       struct nngfs_dir_entry *de,*prevde;
       bh = sb_bread(sb,nip->i_ino+1);
       if(!bh){
          printk(KERN_INFO "NGFS (%s) : Could not read block [%d] \n",__FUNCTION__,inum+1);
          return -1; 
       }
       de = prevde = (struct nngfs_dir_entry *)bh->b_data;
     //  while(strcmp(de->name,"") && ((char *)bh->b_data - (char *)de) <= sb->s_blocksize){
       while(strcmp(de->name,"")){
           if(de->inode == inum){
             de->inode = NNGFS_INUM_INVALID;
           //  printk(KERN_INFO "Setting entry of [%s] as [%d] from [%d] \n",prevde->name,prevde->rec_len,prevde->rec_len+de->rec_len);
             prevde->rec_len +=  de->rec_len;
             mark_buffer_dirty(bh);  
             brelse(bh);
             return 0;
           }
         prevde = de;
         de = (struct nngfs_dir_entry *)((char *)de + de->rec_len);
       }
       brelse(bh);
       return -1;
}

struct inode_operations nngfs_dir_inode_operations ={
        .lookup = nngfs_lookup,
        .create = nngfs_create,
        .unlink = nngfs_unlink,
};

